// Copyright 2018 Google Inc. All Rights Reserved.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS-IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#ifndef S2_UTIL_GTL_LAYOUT_H_
#define S2_UTIL_GTL_LAYOUT_H_

#include "s2/third_party/absl/container/internal/layout.h"

namespace gtl {

using absl::container_internal::Aligned;
using absl::container_internal::Layout;

}  // namespace gtl

#endif  // S2_UTIL_GTL_LAYOUT_H_
