// Copyright 2018 Google Inc. All Rights Reserved.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS-IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#ifndef S2_BASE_SPINLOCK_H_
#define S2_BASE_SPINLOCK_H_

#include <atomic>


class SpinLock {
 public:
  SpinLock() = default;
  ~SpinLock() = default;
  SpinLock(SpinLock const&) = delete;
  SpinLock& operator=(SpinLock const&) = delete;

  inline void Lock() {
    while (flag_.test_and_set(std::memory_order_acquire)) {
      // Spin.
      continue;
    }
  }

  inline void Unlock() {
    flag_.clear(std::memory_order_release);
  }

 private:
  std::atomic_flag flag_ = ATOMIC_FLAG_INIT;
};

#endif  // S2_BASE_SPINLOCK_H_
